"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

from dynaconf.loaders.vault_loader import list_envs
"""
import shortuuid
from dynaconf import settings
from dynaconf.utils.files import SEARCHTREE
from loguru import logger


class Conf:
    """Manage application settings"""
    _version: str
    log_level = "WARNING"

    def __init__(self) -> None:
        """Constructor for Settings"""
        if hasattr(settings, 'LOG_LEVEL'):
            self.log_level = settings.LOG_LEVEL
        self._version = shortuuid.uuid()
        self._settings = settings
        logger.debug(self._settings.dynaconf_banner)
        logger.debug("Learn more at http://github.com/rochacbruno/dynaconf")
        logger.debug(f'dynaconf search tree: {SEARCHTREE}')
        logger.debug(f'dynaconf loaded? {self._settings.configured}')

    @classmethod
    def log_filter(cls, record) -> bool:
        levelno = logger.level(cls.log_level).no
        return record["level"].no >= levelno

    @property
    def version(self) -> str:
        return self._version

    @property
    def settings(self) -> settings:
        return self._settings



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
import json
import sys

import hvac
import requests
from loguru import logger

from .store_engine import StoreEngine
from .store_object import StoreObject


class VaultRESTClient:
    """REST Client commands not available via hvac"""
    """[TODO 5/17/20] ljohnson add hvac.Client to this object"""

    __headers = {'Content-Type': 'application/json',
                 'X-Vault-Token': ''}
    __url: str  #: Vault server URL
    __token: str  #: Auth token
    __mount: str  #: Engine mount path
    __mounts: json

    def __init__(self, url: str, token: str, mount: str) -> None:
        """Constructor for VaultRESTClient"""
        self.__url = url
        self.__token = token
        self.__mount = mount
        self.__headers['X-Vault-Token'] = token

    def _get(self, path: str) -> json:
        """GET REST API wrapper method

            :param path: Vault API to query
            :type path: String
            :return: JSON paylod
        """
        try:
            response = requests.get(self.__url+path, headers=self.__headers)
            response.raise_for_status()

            return json.loads(response.content.decode('utf-8'))

        except requests.exceptions.HTTPError as errh:
            logger.error(f'Http Error: {errh}')
        except requests.exceptions.ConnectionError as errc:
            logger.error(f'Error Connecting: {errc}')
        except requests.exceptions.Timeout as errt:
            logger.error(f'Timeout Error: {errt}')
        except requests.exceptions.RequestException as err:
            logger.error(f'Error: {err}')

    def _put(self, path: str, data: json) -> requests.Response:
        """PUT REST API wrapper method

            :param path: Vault API to change or create
            :type path: String
            :param data: Required request body
            :type data: JSON
            :return: requests.Response object
        """
        try:
            response = requests.put(self.__url+path, headers=self.__headers, data=data)
            response.raise_for_status()

            return response

        except requests.exceptions.HTTPError as errh:
            logger.error(f'Http Error: {errh}')
        except requests.exceptions.ConnectionError as errc:
            logger.error(f'Error Connecting: {errc}')
        except requests.exceptions.Timeout as errt:
            logger.error(f'Timeout Error: {errt}')
        except requests.exceptions.RequestException as err:
            logger.error(f'Error: {err}')

    def get_mounts(self) -> json:
        """Refresh set of mounts from Vault

            :return: JSON
        """
        self.__mounts = self._get("/v1/sys/mounts")
        return self.__mounts

    def new_mount(self, mount: str) -> bool:
        """Will create a vault mount of type k/v V2 if it doesn't exist

            :param mount: Name of the Vault K/V Secret Engine
            :type mount: String
            :return: Boolean
        """
        self.__mounts = self._get("/v1/sys/mounts")
        if mount+"/" not in self.__mounts['data'].keys():
            logger.info(f'Vault mount {self.__url}/v1/sys/mounts/{mount} does not exist, creating')
            self._put(f'/v1/sys/mounts/{mount}', '{"type": "kv-v2"}')
            self.__mounts = self._get("/v1/sys/mounts")
        return mount+"/" in self.__mounts['data'].keys()


class VaultStoreEngine(StoreEngine):
    """Vault implementation of the StoreEngine interface"""
    __vault_url: str
    __vault_token: str
    __vault_client: hvac.Client
    __vault_RESTClient: VaultRESTClient

    def __init__(self, settings) -> None:
        """Constructor for VaultStoreEngine"""
        super().__init__()
        self._settings = settings
        self.__vault_url = settings.VAULT_URL
        self.__vault_token = settings.VAULT_TOKEN
        self.__vault_mount = settings.VAULT_MOUNT
        self.__vault_client = hvac.Client(url=self.__vault_url, token=self.__vault_token)
        self.__vault_RESTClient = VaultRESTClient(self.__vault_url, self.__vault_token, self.__vault_mount)
        logger.debug(f'🔐 Vault backend configuration loaded. {self.__vault_url}')
        if self.initialize():
            logger.debug("🔐 Vault backend initialized.")

    def initialize(self) -> bool:
        try:
            self.__vault_RESTClient.new_mount(self.__vault_mount)
            self.__vault_client.secrets.kv.configure(
                max_versions=self._settings.VAULT_CLIENT_MAX_VERSIONS,
                cas_required=self._settings.VAULT_CLIENT_CAS,
                mount_point=self.__vault_mount
            )

        except Exception as err:
            logger.error(f'Failed to configure vault client {err}')
            sys.exit(1)

        return self.open()

    def open(self) -> bool:
        """Ensure the vault client is connected"""
        return self.__vault_client.is_authenticated()

    def close(self) -> bool:
        """Ensure we close the vault connection"""
        return self.__vault_client.logout()

    def __del__(self):
        """When the object is destroyed make sure we close the connection to Vault"""
        return self.close()

    def write(self, obj: StoreObject) -> bool:
        """Given a StoreObject, store it into vault using mount/path/name == body,info

            :param obj: The StoreObject to persist
            :type obj: StoreObject
            :return: bool
        """
        client = self.__vault_client
        mp = self.__vault_mount

        try:
            client.secrets.kv.v2.create_or_update_secret(path=obj.path_name + "/cert_body",
                                                         mount_point=mp,
                                                         secret=obj.data['cert_body'])

            client.secrets.kv.v2.create_or_update_secret(path=obj.path_name + "/cert_info",
                                                         mount_point=mp,
                                                         secret=obj.data['cert_info'])
        except Exception as vex:
            logger.error(f'Failed to write StoreObject to Vault {vex}')

        logger.info(f'Successfully stored {obj.path_name}')

    def read(self, path: str, name: str) -> StoreObject:
        """Using the provided path and name retrieve the data from the store and create a new StoreObject

            :param path: Store path to the object
            :type path: str
            :param name: Name of the object to retrieve
            :type name: str
            :return: StoreObject
        """
        client = self.__vault_client
        mp = self.__vault_mount
        cert: StoreObject

        try:
            certbody = client.secrets.kv.v2.read_secret_version(path=path + "/" + name + "/cert_body", mount_point=mp)
            certinfo = client.secrets.kv.v2.read_secret_version(path=path + "/" + name + "/cert_info", mount_point=mp)
            cert = StoreObject(name=name, path=path, body=certbody['data']['data'], info=certinfo['data']['data'])
            cert._data = {'cert_body': certbody['data']['data'], 'cert_info': certinfo['data']['data']}

        except Exception as vex:
            logger.error(f'Failed to read StoreObject /{self.__vault_mount}{path}/{name} {vex}')

        logger.info(f'Successfully read {cert.path_name}')

        return cert



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
import hashlib
import json


class StoreObject:
    """Metadata interface for objects being persisted in a backend"""
    _name: str  #: Name of the objects store key
    _path: str  #: Path from store mount point to find store key
    _body: str  #: Content that will be persisted
    _info: str  #: Metadata about the object being stored
    _data: {}   #: Complete map of object
    _version: int  #: Store revision

    def __init__(self, name: str, path: str, body: str, info: str) -> None:
        """Constructor for StoreObject"""

        self._name = name  #: [TODO 5/16/20] ljohnson name validator
        self._path = path  #: [TODO 5/16/20] ljohnson path validator
        self._body = body  #: [TODO 5/16/20] ljohnson body and info validator
        self._info = info

    @property
    def name(self) -> str:
        """Object name"""
        return self._name

    @property
    def path(self) -> str:
        """Path attribute"""
        return self._path

    @property
    def path_name(self) -> str:
        """Convienence method to generate path/name for store"""
        return self._path + "/" + self._name

    @property
    def body(self) -> str:
        """Content to persist, typically JSON"""
        return self._body

    @property
    def info(self) -> str:
        """Object metadata"""
        return self._info

    @property
    def data(self) -> {}:
        return self._data

    def md5(self) -> str:  # noqa F811
        return hashlib.md5(json.dumps(self._data, sort_keys=True).encode('utf-8')).hexdigest()

    @staticmethod  # noqa F811
    def md5(obj: {}) -> str:
        return hashlib.md5(json.dumps(obj, sort_keys=True).encode('utf-8')).hexdigest()

    @property
    def version(self) -> int:
        """Object version"""
        return self._version

    @name.setter
    def name(self, value: str) -> None:
        self._name = value

    @path.setter
    def path(self, value: str) -> None:
        self._path = value

    @body.setter
    def body(self, value: str) -> None:
        self._body = value

    @info.setter
    def info(self, value: str) -> None:
        self._info = value

    @version.setter
    def version(self, value: int) -> None:
        self._version = value



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
from loguru import logger

from .store_engine import StoreEngine
from .store_file import FileStoreEngine
from .store_object import StoreObject
from .store_vault import VaultStoreEngine


class Store:
    """Abstract class to generalize access to the different stores"""
    _engine: StoreEngine
    _engine_map = {
        'vault': VaultStoreEngine,
        'file': FileStoreEngine
    }

    def __init__(self, settings) -> None:
        """Dynamically load StoreEngine type from .env via Dynaconf
        KNOX_STORE_ENGINE=[vault,file]

        :param settings: Dynaconf LazySettings
        :type settings: dynaconf.LazySettings
        """
        try:
            self._engine = self._engine_map.get(settings.STORE_ENGINE).__call__(settings)
        except Exception:
            logger.error(f'StoreEngineFailure KNOX_STORE_ENGINE={settings.STORE_ENGINE} is invalid. Valid options are {self._engine_map.keys()}')  # noqa: E501

        self._engine.settings = settings
        logger.debug(f'Loaded {self._engine.__class__}')

    def save(self, obj: StoreObject) -> bool:
        """Save the given object to persistence"""
        return self._engine.write(obj)

    def get(self, path: str, name: str) -> StoreObject:
        """Given path read object"""
        return self._engine.read(path, name)

    def delete(self, path: str, name: str) -> bool:
        """Remove the object from the store"""
        """[TODO 5/13/20] ljohnson implement soft delete and hard deletes"""
        return self._engine.delete(path, name)

    def find(self, path: str, name: str) -> [StoreObject]:
        """Given a path, return collection of all objects"""
        """[TODO 5/13/20] ljohnson implement glob searching /path/**/path2/*"""
        pass



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
from .store import Store  # noqa: F401
from .store_engine import StoreEngine  # noqa: F401
from .store_file import FileStoreEngine  # noqa: F401
from .store_object import StoreObject  # noqa: F401
from .store_vault import VaultRESTClient  # noqa: F401
from .store_vault import VaultStoreEngine  # noqa: F401



"""
Apache Software License 2.0

Copyright (c) 2020, 8x8, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License."""
import os  # noqa: F401

from loguru import logger

from .store_engine import StoreEngine


class FileStoreEngine(StoreEngine):
    """"""
    __file_home: str

    def __init__(self, settings) -> None:
        """Constructor for FileStore"""
        super().__init__()
        self.__file_home = settings.FILE_HOME
        logger.debug(f'📂 File backend configuration loaded. {self.__file_home}')
